#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/printk.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/of_irq.h>
#include <linux/module.h>
#include <linux/moduleparam.h>

#include <rbus/reg.h>

static void __iomem *iobase;

irqreturn_t isr_sb2_dbg(int irq, void *reg_base)
{
	struct pt_regs *regs;
	u32 intr = readl(SB2_DBG_INT);

	regs = get_irq_regs();

	pr_info("sb2 get int 0x%08x\n", intr);

	if (intr & ((1 << 4) | (1 << 10) | (1<<6) | (1<<12))) {
		u32 cause, addr, s_a_cpu;
		char buf[128];

		writel((1 << 9) | (1 << 7) | 1, SB2_DBG_INT);

		s_a_cpu = (intr & (1<<10)) ? 1 : 2;	/* SCPU:1, ACPU:2 */
		addr = (s_a_cpu==1) ? readl(SB2_DBG_ADDR_SYSTEM) : readl(SB2_DBG_ADDR_AUDIO);
		cause = readl(SB2_DBG_ADDR1);
		cause = (s_a_cpu==1) ? (cause >> 2) : (cause >> 4);

		sprintf(buf, "Memory 0x%08x trashed by %sCPU with %s %s\n", addr,
				(s_a_cpu == 1) ? "S" : "A",
				(cause & 1) ? "D" : "I", 
				(cause & 2) ? "W" : "R");

		die(buf, regs, 0);

		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}

static const struct of_device_id rtk_sb2_match[] __initconst = {
	{.compatible = "Realtek,rtk-sb2"},
	{},
};

static int __init sb2_dbg_init(void)
{
	struct device_node *np;
	int irq;

	pr_info("sb2_dbg_init\n");

	np = of_find_matching_node(NULL, rtk_sb2_match);
	if (unlikely(np == NULL))
		return -1;

	irq = irq_of_parse_and_map(np, 0);

	iobase = of_iomap(np, 0);
	/* pr_info("sb2 irq %d, iobase 0x%08x\n", irq, (u32)iobase); */

	of_node_put(np);

	if (request_irq(irq, isr_sb2_dbg, IRQF_SHARED, "sb2", iobase) != 0) {
		pr_err("Cannot get IRQ\n");
		return -1;
	}

	writel((1 << 9) | (1 << 7) | 1, SB2_DBG_INT);

	return 0;
}
late_initcall(sb2_dbg_init);

static void __exit sb2_dbg_exit(void)
{
//	struct sb2_dbg_info *info = &test_info;
//printk("%s:%d ===> \n", __func__, __LINE__);
	struct device_node *np;
	int irq;

	printk("sb2_dbg_exit\n");

	np = of_find_matching_node(NULL, rtk_sb2_match);
	if (unlikely(np == NULL))
		return -1;

	irq = irq_of_parse_and_map(np, 0);

	iobase = of_iomap(np, 0);
	/* pr_info("sb2 irq %d, iobase 0x%08x\n", irq, (u32)iobase); */
	printk("sb2 irq %d, iobase 0x%08x\n", irq, (u32)iobase);

	of_node_put(np);
	free_irq(irq, NULL);
//	debugfs_remove_recursive(info->root);
//	stop_threaded_test(info);
//	result_free(info, NULL);
}
module_exit(sb2_dbg_exit);


void sb2_dbg_disable_mem_monitor(int which)
{
	u32 *reg_ctrl = (u32 *) SB2_DBG_CTRL_REG0;

	reg_ctrl += which;
	writel((1 << 13) | (1 << 9) | (1 << 1), reg_ctrl);
}
EXPORT_SYMBOL(sb2_dbg_disable_mem_monitor);

static void sb2_dbg_set_mem_monitor(int which, u32 start, u32 end, u32 flags)
{
	u32 *reg_start, *reg_end, *reg_ctrl;

	/* disable this set first */
	sb2_dbg_disable_mem_monitor(which);

	reg_start = ((u32 *) SB2_DBG_START_REG0) + which;
	reg_end = ((u32 *) SB2_DBG_END_REG0) + which;
	reg_ctrl = ((u32 *) SB2_DBG_CTRL_REG0) + which;

	writel(start, reg_start);
	writel(end, reg_end);

	writel(flags, reg_ctrl);

	/*
	pr_info("sb2 0x%08x=0x%08x\n", (u32)reg_ctrl, readl(reg_ctrl));
	pr_info("sb2 0x%08x=0x%08x\n", (u32)reg_start, readl(reg_start));
	pr_info("sb2 0x%08x=0x%08x\n", (u32)reg_end, readl(reg_end));
	pr_info("sb2 0x%08x=0x%08x\n", (u32)SB2_DBG_INT, readl(SB2_DBG_INT));
	*/
}

/*
 * which: 0~7, which register set
 * d_i: SB2_DBG_MONITOR_DATA|SB2_DBG_MONITOR_INST
 * r_w: SB2_DBG_MONITOR_READ|SB2_DBG_MONITOR_WRITE
 */
void sb2_dbg_scpu_monitor(int which, u32 start, u32 end, u32 d_i, u32 r_w)
{
	sb2_dbg_set_mem_monitor(which, start, end, (1 << 13) | (3 << 8) | d_i | r_w | 3);
}
EXPORT_SYMBOL(sb2_dbg_scpu_monitor);

void sb2_dbg_acpu_monitor(int which, u32 start, u32 end, u32 d_i, u32 r_w)
{
	sb2_dbg_set_mem_monitor(which, start, end, (3 << 12) | (1 << 9) | d_i | r_w | 3);
}
EXPORT_SYMBOL(sb2_dbg_acpu_monitor);
