/*
 * drivers/pwm/pwm-rtd1195.c
 *
 * rtd1195 pulse-width-modulation controller driver
 *
 * Copyright (c) 2014, Realtek Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/clk.h>
#include <linux/err.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/pwm.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/sysfs.h>

// RTK define here
#define RTK_DEF_PWM_IDX     (0)     // default use PWM idx set to 0
#define RTK_DEF_GPIO_IDX    (11)    // default use GPIO pin set to 11

#define RTK_ADDR_MUXPAD0    (0x18007310)
#define RTK_ADDR_PWM_OCD    (0x180070D0)
#define RTK_ADDR_PWM_CD     (RTK_ADDR_PWM_OCD+4)
#define RTK_ADDR_PWM_CSD    (RTK_ADDR_PWM_OCD+8)

#define RTK_PWM_OCD_SHIFT       (8)
#define RTK_PWM_CD_SHIFT        (8)
#define RTK_PWM_CSD_SHIFT       (4)

#define RTK_PWM_OCD_DEFAULT      	(0x1)
#define RTK_PWM_CD_DEFAULT       	(0x1)
#define RTK_PWM_CSD_DEFAULT     	(0x8)

#define NUM_PWM 4


struct RTK119X_pwm_map {
	int	duty_rate;
	int ocd_data;
	int cd_data;
};

static const struct RTK119X_pwm_map rtk119x_maps[] = {
	{100, 1, 1}, // 1
	{80, 4, 3},  // 4/5
	{75, 3, 2},  // 3/4
	{66, 2, 1},  // 2/3
	{60, 4, 2},  // 3/5
	{50, 3, 1},  // 1/2
	{40, 4, 1},	 // 2/5
	{33, 5, 1},	 // 1/3
	{25, 3, 0},	 // 1/4
	{20, 3, 0},	 // 1/5
	{0, 0, 0},   // 0
};

struct rtd1195_pwm_chip {
	struct pwm_chip		chip;
	struct device		*dev;
	void __iomem		*mmio_base;
	void __iomem		*mmio_pwm_reg_base;	
	int					proc_id;	
	int                 pwm_idx;
	int     			duty_rate;
	int     	        enable;
	int                 clksrc_div;
};

ssize_t pwm_show_dutyRate(struct device *dev, struct device_attribute *attr, char *buf);
ssize_t pwm_store_dutyRate(struct device *dev, struct device_attribute *attr,
			 const char *buf, size_t count);

ssize_t pwm_show_enable(struct device *dev, struct device_attribute *attr, char *buf);
ssize_t pwm_store_enable(struct device *dev, struct device_attribute *attr,
			 const char *buf, size_t count);

ssize_t pwm_show_clkDiv(struct device *dev, struct device_attribute *attr, char *buf);
ssize_t pwm_store_clkDiv(struct device *dev, struct device_attribute *attr,
			 const char *buf, size_t count);


static DEVICE_ATTR(dutyRate, S_IRWXUGO, pwm_show_dutyRate, pwm_store_dutyRate);
static DEVICE_ATTR(enable, S_IRWXUGO, pwm_show_enable, pwm_store_enable);
static DEVICE_ATTR(clkDiv, S_IRWXUGO, pwm_show_clkDiv, pwm_store_clkDiv);


static struct attribute *pwm_dev_attrs[] = {
	&dev_attr_dutyRate.attr,
	&dev_attr_enable.attr,
	&dev_attr_clkDiv.attr,
	NULL,
};

static struct attribute_group pwm_dev_attr_group = {
	.attrs		= pwm_dev_attrs,
};

static inline struct rtd1195_pwm_chip *to_rtd1195_pwm_chip(struct pwm_chip *chip)
{
	return container_of(chip, struct rtd1195_pwm_chip, chip);
}

#define to_pwm_device(d) return container_of(d, struct rtd1195_pwm_chip, chip);

static int pwm_map_idx_lookup(int duty_rate) {
	int idx=0;
	int duty=duty_rate;
	int found = 0;
	int mapSize=ARRAY_SIZE(rtk119x_maps);
	
	if( duty >= 100) {
		duty = 100;
		idx=0;
		pr_info("--- debug : pwm_map_idx_lookup, duty_rate as (%d) is set to 100, idx=%d \n", duty, idx);		
		return idx;
	}
	else if (duty <= 0 ) {
		duty = 0;
		idx=mapSize-1;
		pr_info("--- debug : pwm_map_idx_lookup, duty_rate as (%d) is set to 0, idx=%d \n", duty, idx);		
		return idx;
	}

	idx=0;
	while(!found) {
		if( idx==(mapSize-1) )
			break;
		if( duty <= rtk119x_maps[idx].duty_rate && duty > rtk119x_maps[idx+1].duty_rate ) {
			found = 1;
		}
		else {
			idx++;
		}
	}
	pr_info("--- debug : pwm_map_idx_lookup, duty_rate as (%d) , found idx=%d \n", duty, idx);		
	return idx;
}

//static void pwm_set_register()
static void pwm_set_register(struct rtd1195_pwm_chip *pc)
{
	int map_idx = 0;

	if(pc==NULL) {
		pr_err("--- debug : pwm_set_register -- parameter error!-- \n");
		return;
	}
	
	map_idx = pwm_map_idx_lookup(pc->duty_rate);	
	pr_info("--- debug : pwm_set_register -- map_idx=%d set OCD=%d,CD=%d,CSD=%d,duty_rate=%d \n", map_idx, 
	rtk119x_maps[map_idx].ocd_data, rtk119x_maps[map_idx].cd_data, pc->clksrc_div, pc->duty_rate); 
	
	writel(rtk119x_maps[map_idx].ocd_data << (pc->pwm_idx*RTK_PWM_OCD_SHIFT), pc->mmio_pwm_reg_base);
	writel(rtk119x_maps[map_idx].cd_data << (pc->pwm_idx*RTK_PWM_CD_SHIFT), (pc->mmio_pwm_reg_base+4));
	writel(pc->clksrc_div << (pc->pwm_idx*RTK_PWM_CSD_SHIFT), (pc->mmio_pwm_reg_base+8)); 
	pr_info("--- debug :  ---- pwm_set_register --- done! \n");
}

//static void pwm_reset_register()
static void pwm_reset_register(struct rtd1195_pwm_chip *pc)
{
	if(pc==NULL) {
		pr_err("--- debug : pwm_reset_register -- parameter error!-- \n");
		return;
	}

	writel(0, (pc->mmio_pwm_reg_base));
	writel(0, (pc->mmio_pwm_reg_base+4));
	writel(0, (pc->mmio_pwm_reg_base+8)); 
	pr_info("--- debug : pwm_reset_register --Done!-- \n");
}

static int rtd1195_pwm_config(struct pwm_chip *chip, struct pwm_device *pwm,
			    int duty_ns, int period_ns)
{
	struct rtd1195_pwm_chip *pc = to_rtd1195_pwm_chip(chip);
	pr_info("--- debug : rtd1195_pwm_config  : enable=%d \n",pc->enable);
	if( pc->enable == 0 ) {
		pwm_reset_register(pc);
	}
	else {
		pwm_set_register(pc);
	}
	return 0;
}

static int rtd1195_pwm_enable(struct pwm_chip *chip, struct pwm_device *pwm)
{
	struct rtd1195_pwm_chip *pc = to_rtd1195_pwm_chip(chip);
	pr_info("--- debug : rtd1195_pwm_enable ---- duty_rate=%d \n", pc->enable);
	if( pc->enable == 0 ) {
		pwm_reset_register(pc);
	}	
	return 0;
}

static void rtd1195_pwm_disable(struct pwm_chip *chip, struct pwm_device *pwm)
{
	struct rtd1195_pwm_chip *pc = to_rtd1195_pwm_chip(chip);
	pr_info("--- debug :  rtd1195_pwm_disable ---- disable PWM now !\n");
	pwm_reset_register(pc);
}

static const struct pwm_ops rtd1195_pwm_ops = {
	.config = rtd1195_pwm_config,
	.enable = rtd1195_pwm_enable,
	.disable = rtd1195_pwm_disable,
	.owner = THIS_MODULE,
};

/** define show/store API for each file here **/
ssize_t pwm_show_dutyRate(struct device *dev, struct device_attribute *attr, char *buf)
{	
	struct platform_device *pdev = to_platform_device(dev);		
	struct rtd1195_pwm_chip *pc = platform_get_drvdata(pdev);	

	pr_info("--- debug : pwm_show_dutyRate called : dutyRate=> pc->dutyRate=[%d]--- \n", pc->duty_rate);

	return sprintf(buf, "%d\n", pc->duty_rate);
}

ssize_t pwm_store_dutyRate(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	int value=0;
	int ret=-1;
	struct platform_device *pdev = to_platform_device(dev);		
	struct rtd1195_pwm_chip *pc = platform_get_drvdata(pdev);	

	pr_info("--- debug : pwm_store_dutyRate called : count:%d ",count);
	
	if( count < 1 ) {
		pr_err("--- debug : pwm_store_dutyRate , count is too small, return \n");
		return count;
	} 
	ret = kstrtoint(buf, 10, &value);
	if ( ret != 0 ) {
		pr_err("--- debug : pwm_store_dutyRate , parse buf error! ret=%d \n", ret);
		return count;
	}
	pr_info(" value:[%d] \n",value);
	if( value < 0 || value > 100 ) {
		pr_err("--- debug : pwm_store_dutyRate  ====  input should betweek 0 ~ 100 \n");
		return count;
	}
	if( pc->duty_rate == value ) {
		pr_err(" debug : pwm_store_dutyRate -- dutyRate value is not change, return! \n");
		return count;
	}
	else {
		pr_err(" debug : pwm_store_dutyRate -- assign [%d] to duty_rate now! \n", value);
		pc->duty_rate = value;
		if( pc->enable == 1 )
			pwm_set_register(pc);
	}
	
	return count;
}
			 
ssize_t pwm_show_enable(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct platform_device *pdev = to_platform_device(dev);		
	struct rtd1195_pwm_chip *pc = platform_get_drvdata(pdev);	

	pr_info("--- debug : pwm_show_enable called : enable=%d---- \n", pc->enable);

	return sprintf(buf, "%d\n", pc->enable);
}

ssize_t pwm_store_enable(struct device *dev, struct device_attribute *attr,
			 const char *buf, size_t count)	 
{
	struct platform_device *pdev = to_platform_device(dev);		
	struct rtd1195_pwm_chip *pc = platform_get_drvdata(pdev);	
	int value=0;
	
	if(buf ==NULL) {
		pr_err("--- debug : rtd1195_pwm_enable	====  buffer is null, return \n");
		return count;
	} 
	sscanf(buf, "%d", &value);

	if( pc->enable == value ) {
		pr_err("--- debug : rtd1195_pwm_enable	====  the same, do nothing (value=%d) \n", value);
		return count;
	}

	pc->enable = value;
	if(value==1) {
		pwm_set_register(pc);
	}
	else {
		pwm_reset_register(pc);
	}
	pr_info("--- debug : rtd1195_pwm_enable  ====  done \n");
	return count;
}

ssize_t pwm_show_clkDiv(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct platform_device *pdev = to_platform_device(dev);		
	struct rtd1195_pwm_chip *pc = platform_get_drvdata(pdev);	

	pr_info("--- debug : pwm_show_clkDiv called :clksrc_div=%d---- \n", pc->clksrc_div);

	return sprintf(buf, "%d\n", pc->clksrc_div);
}

ssize_t pwm_store_clkDiv(struct device *dev, struct device_attribute *attr,
			 const char *buf, size_t count)	 
{
	struct platform_device *pdev = to_platform_device(dev);		
	struct rtd1195_pwm_chip *pc = platform_get_drvdata(pdev);	

	int value=0;
	if(buf ==NULL) {
		pr_err("--- debug : pwm_store_clkDiv	====  buffer is null, return \n");
		return count;
	} 
	sscanf(buf, "%d", &value);

	if(value < 0 || value > 15) {
		pr_err("--- debug : pwm_store_clkDiv  ====  input should betweek 0 ~ 15 \n");
		return count;
	}

	if(pc->clksrc_div == value) {
		pr_err("--- debug : pwm_store_clkDiv  ====  input is the same=%d, do nothing! \n", value);
		return count;
	}

	pc->clksrc_div = value;
	if( pc->enable == 1 )
		pwm_set_register(pc);

	pr_info("--- debug : pwm_store_clkDiv  ====  done (value=%d) \n", value);
	return count;
}

static int rtd1195_pwm_probe(struct platform_device *pdev)
{
	struct rtd1195_pwm_chip *pwm;
	int ret=0;
	u32 val=0;
	pr_info("--- debug : rtd1195_pwm_probe ---- \n");

	pwm = devm_kzalloc(&pdev->dev, sizeof(*pwm), GFP_KERNEL);
	if (!pwm) {
		dev_err(&pdev->dev, "failed to allocate memory\n");
		return -ENOMEM;
	}
	pwm->proc_id = pdev->id;
	pwm->dev = &pdev->dev;
	
	platform_set_drvdata(pdev, pwm);
	
	pwm->mmio_base = ioremap(RTK_ADDR_MUXPAD0, 4);
	pwm->mmio_pwm_reg_base =  ioremap(RTK_ADDR_PWM_OCD, 4*3);
	
	if (IS_ERR(pwm->mmio_base))
		return PTR_ERR(pwm->mmio_base);

	pwm->enable=0;
	pwm->clksrc_div=0;
	pwm->duty_rate=0;
	pwm->pwm_idx=0;

	if(!of_property_read_u32(pdev->dev.of_node, "duty_rate", &val))
		pwm->duty_rate=val;
	
	if (!of_property_read_u32(pdev->dev.of_node, "clksrc_div", &val))
		pwm->clksrc_div=val;
	
	if (!of_property_read_u32(pdev->dev.of_node, "enable", &val))
		pwm->enable=val;
	
	if(!of_property_read_u32(pdev->dev.of_node, "pwm_idx", &val))
		pwm->pwm_idx=val;
			
	pr_err("--- debug : rtd1195_pwm_probe - enable=(%d) duty_rate=(%d) clksrc_div==(%d)---  \n",pwm->enable, pwm->duty_rate, pwm->clksrc_div);
    

	pwm->chip.dev = &pdev->dev;
	pwm->chip.ops = &rtd1195_pwm_ops;
	pwm->chip.base = 0; // -1;
	pwm->chip.npwm = NUM_PWM;
	pwm->chip.can_sleep = true;

	ret = sysfs_create_group(&pdev->dev.kobj, &pwm_dev_attr_group);
	if (ret < 0) {
		dev_err(&pdev->dev, "sysfs_create_group() failed: %d\n", ret);
		return ret;
	}

	ret = pwmchip_add(&pwm->chip);
	if (ret < 0) {
		dev_err(&pdev->dev, "pwmchip_add() failed: %d\n", ret);
		sysfs_remove_group(&pdev->dev.kobj, &pwm_dev_attr_group);
		return ret;
	}

	val = readl(pwm->mmio_base);
	pr_err("--- debug : rtd1195_pwm_probe ---- set PWM[%d] out , read = (0x%x) !\n",pwm->pwm_idx, val); 

    if( pwm->enable == 0 ) {
        pwm_reset_register(pwm);
    }
    else {
        pwm_set_register(pwm);
    }
	pr_err("--- debug : rtd1195_pwm_probe --  done! !\n"); 

	return 0;
}

static int rtd1195_pwm_remove(struct platform_device *pdev)
{
	struct rtd1195_pwm_chip *pc = platform_get_drvdata(pdev);

	pwm_reset_register(pc);
	sysfs_remove_group(&pdev->dev.kobj, &pwm_dev_attr_group);
	
	pr_err("--- debug : rtd1195_pwm_remove ---- \n");

	return pwmchip_remove(&pc->chip);
}

static const struct of_device_id rtd1195_pwm_of_match[] = {
	{ .compatible = "realtek,rtd1195-pwm" },
	{ }
};
MODULE_DEVICE_TABLE(of, rtd1195_pwm_of_match);

static struct platform_driver rtd1195_pwm_platform_driver = {
	.driver = {
		.name = "pwm",
		.of_match_table = rtd1195_pwm_of_match,
	},
	.probe = rtd1195_pwm_probe,
	.remove = rtd1195_pwm_remove,
};
module_platform_driver(rtd1195_pwm_platform_driver);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("REALTEK Corporation");
MODULE_ALIAS("platform:rtd1195-pwm");
